import os, sys
from importlib import import_module
from inspect import signature, Parameter
from functools import partial, wraps
from typing import List, get_origin, get_args
from types import GenericAlias
from framework.base import get_project_path

__COMPONENT_CLASS = []
__NAME_COMPONENT = {}
__COMPONENT_INSTANCE = {}


class InjectableError(Exception):
    def __init__(self, msg):
        self.__msg = msg

    def __str__(self):
        return self.__msg


def component(cls=None, bean_name=None):
    if cls is None:
        return partial(component, bean_name=bean_name)

    __COMPONENT_CLASS.append(cls)
    if bean_name:
        if bean_name in __NAME_COMPONENT:
            raise InjectableError(f"bean name[{bean_name}] is already defined.")
        __NAME_COMPONENT[bean_name] = cls
    return cls


def autowired(func=None, names=None):
    if func is None:
        return partial(autowired, names=names)

    @wraps(func)
    def wrap(*args, **kwargs):
        if func.__name__ == '__init__':
            self_type = type(args[0])
            if self_type in __COMPONENT_CLASS and self_type not in __COMPONENT_INSTANCE:
                __COMPONENT_INSTANCE[self_type] = args[0]
        for p_name, p_type in signature(func).parameters.items():
            if p_name == 'self' or p_type == Parameter.empty or p_name in kwargs:
                continue
            if names is not None and p_name in names:
                if names[p_name] in __NAME_COMPONENT:
                    kwargs[p_name] = get_instance(__NAME_COMPONENT[names[p_name]])
            elif get_origin(p_type.annotation) is list:
                inject_types = get_args(p_type.annotation)
                if len(inject_types) > 0:
                    instances = [get_instance(x) for x in __COMPONENT_CLASS if issubclass(x, inject_types)]
                    if len(instances) > 0:
                        kwargs[p_name] = instances
            else:
                components = [x for x in __COMPONENT_CLASS if issubclass(x, p_type.annotation)]
                if len(components) > 0:
                    kwargs[p_name] = get_instance(components[0])
        func(*args, **kwargs)

    return wrap


def get_instance(t):
    if t not in __COMPONENT_CLASS:
        return None
    if t not in __COMPONENT_INSTANCE:
        __COMPONENT_INSTANCE[t] = t()
    return __COMPONENT_INSTANCE[t]


def init_injectable(path=get_project_path()):
    for f in os.listdir(path):
        if os.path.isdir(f) and os.path.exists(os.path.join(f, '__init__.py')):
            init_injectable(path=os.path.join(path, f))
        if f.endswith('.py') and f != '__init__.py':
            py = os.path.relpath(os.path.join(path, f), get_project_path())[:-3]
            import_module('.'.join(py.split(os.path.sep)))
